import sys
import z3
from pdb import pm
from miasm2.analysis.machine import Machine
from miasm2.analysis.binary import Container
from miasm2.ir.symbexec import symbexec
from collections import namedtuple
from miasm2.ir.translators import Translator
from miasm2.expression.expression import *
from miasm2.expression.expression_helper import possible_values

cont = Container.from_stream(open(sys.argv[1]))
machine = Machine(cont.arch)
mdis = machine.dis_engine(cont.bin_stream)
blocks = mdis.dis_multibloc(cont.entry_point)
arg1 = ExprId("arg1")
trans = Translator.to_language("z3")

ira = machine.ira()
for block in blocks:
    ira.add_bloc(block)

State = namedtuple("State", ["addr", "symbexec", "constraints"])

class SbExecStopINT(symbexec):

    def emul_ir_blocs(self, myir, ad, lbl_stop=None, step=False):
        while True:
            if (isinstance(self.symbols[myir.arch.regs.interrupt_num], ExprInt) and
                int(self.symbols[myir.arch.regs.interrupt_num].arg) == 0x80):
                return False
            b = myir.get_bloc(ad)
            if b is None:
                break
            if b.label == lbl_stop:
                break
            ad = self.emulbloc(b, step=step)
        return ad

def get_char(model, idx, trans):
    """Retrieve the input character number @idx of solution @model"""
    addr = ExprMem(arg1 + ExprInt32(idx), 8)
    return chr(model.eval(trans.from_expr(addr)).as_long())

def check_for_solution(sbcpy):
    # Get the message printed by sys_write
    msg_addr = int(sbcpy.symbols[ira.arch.regs.ECX].arg)
    lgth = int(sbcpy.symbols[ira.arch.regs.EDX].arg)
    msg = cont.executable.virt.get(msg_addr, msg_addr + lgth).strip()
    print "REACH INT, message: %s" % msg

    if msg_addr == 0x804B30C:
        # Solution found if message is "Well done!"
        print "SOLUTION FOUND:",

        # Model the solution
        mod = solver.model()

        # Print solution
        cur_char, idx, out = get_char(mod, 0, trans), 0, []
        while cur_char != "\x00":
            out.append(cur_char)
            idx += 1
            cur_char = get_char(mod, idx, trans)
        print "".join(out)


symb = SbExecStopINT(ira, ira.arch.regs.regs_init)
symb.symbols[ExprMem(ExprId("ESP_init") + ExprInt32(8), 32)] = arg1
init_state = State(cont.entry_point, symb, frozenset())
todo = set([init_state])
done = set()

while todo:
    state = todo.pop()
    if state in done:
        continue
    done.add(state)

    # Check reachability
    print "Check %s" % state.addr
    solver = z3.Solver()
    for cons in state.constraints:
        solver.add(trans.from_expr(cons))
    if solver.check().r == z3.unsat:
        # Unreachable
        continue

    sbcpy = SbExecStopINT(ira, {})
    sbcpy.symbols = state.symbexec.symbols.copy()

    addr = sbcpy.emul_ir_blocs(ira, state.addr)

    if addr is False:
        check_for_solution(sbcpy)
        continue

    for out in possible_values(addr):
        constraints = set(state.constraints)
        constraints.update(constraint.to_constraint()
                           for constraint in out.constraints)
        todo.add(State(out.value, sbcpy, frozenset(constraints)))
